/*
 * main.c
 *
 *  Created on: 13.09.2009
 *      Author: christian-g (im) mikrocontroller.net
 *
 *      init and softSPI heavily based on:
 *      http://home.arcor.de/wehrsdorf/Oled-Display-Recycling.html
 */



#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include "bitmaps.h"
#include <compat/deprecated.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include <stdlib.h>


/* !!! für Software-SPI (original) hier eine Null eintragen */
#define HW_SPI		0	// 1=hardware SPI


#define TSLEEP	30	// geht in Sleep-Mode nach TSLEEP Sekunden

#define MAXSPRITES	8	// max. Anzahl Sprites
#define MAXPATTERN	8	// max. Anzahl Spritepattern

#define OLED_12V_HI sbi( PORTD, PORTD5) //PORTD.5
#define OLED_12V_LO cbi( PORTD, PORTD5) //PORTD.5

#define OLED_SDIN_HI  sbi( PORTD, PORTD1) //PORTD:xxxx.xxIx
#define OLED_SDIN_LO  cbi( PORTD, PORTD1);

#define OLED_SCLK_HI  sbi( PORTD, PORTD4); //PORTD:xxxx.xxIx
#define OLED_SCLK_LO  cbi( PORTD, PORTD4);

#define OLED_DC_HI  sbi( PORTC, PORTC1); //PORTD:xxxx.xxIx
#define OLED_DC_LO  cbi( PORTC, PORTC1);

#define OLED_RES_HI  sbi( PORTC, PORTC2); //PORTD:xxxx.xxIx
#define OLED_RES_LO  cbi( PORTC, PORTC2);

#define OLED_CS_HI  sbi( PORTB, PORTB1); //PORTD:xxxx.xxIx
#define OLED_CS_LO  cbi( PORTB, PORTB1);

#define delay_ms	_delay_ms
#define delay_us	_delay_us

#define CLS		drawRect(7,0,57,79,0,1);_delay_us(400);

/* ----------------------------------------------------------------------------------*/

typedef uint8_t uint8;

typedef struct {uint8 x; uint8 y; uint8 scol; uint8 srow;
				uint8 ecol; uint8 erow; uint8 bcol; uint8 brow;
				uint8 op; uint8 col; uint8 po; uint8 dx; uint8 dy;
}Sprite;

// Zeiger auf Spritepattern und Puffer, vgl. doSprite()
struct Pattern {uint8 scol; uint8 srow;	uint8 ecol; uint8 erow;
				uint8 bcol; uint8 brow;
} pt[] = {
			{ 0x3a, 0,  0x3d,  6, 0x42, 28 },
			{ 0x3e, 0,  0x41,  6, 0x42, 35 },
			{ 0x42, 0,  0x45,  6, 0x3a, 42 },
			{ 0x3a, 14, 0x3d, 20, 0x3e, 42 },
			{ 0x3e, 14, 0x41, 20, 0x42, 42 },
			{ 0x42, 14, 0x45, 20, 0x3a, 49 },
			{ 0x3a, 28, 0x3d, 34, 0x3e, 49 },
			{ 0x3e, 28, 0x41, 34, 0x42, 49 }
};

// Zeiger auf Fonts für Zahlen
struct Digits {uint8 scol; uint8 srow;} digit[] = {
			{ 0x3d, 0x49 },	// 0
			{ 0x3a, 0x3b }, // 1
			{ 0x3d, 0x3b }, // 2
			{ 0x40, 0x3b },	//
			{ 0x43, 0x3b },
			{ 0x3a, 0x42 },
			{ 0x3d, 0x42 },
			{ 0x40, 0x42 },
			{ 0x43, 0x42 }, // 8
			{ 0x3a, 0x49 }	// 9
};


void write_byte(uint8 dat_or_cmd);
void write_cmd(uint8 command);
void write_dat(uint8 data);

void delay(uint8);
void OledInit(void);
void Demo(void);
void doSprite(Sprite *sprite);
void drawRect(uint8 scol, uint8 srow, uint8 ecol, uint8 erow, uint8 pattern, uint8 fill);
void doCopy(uint8 scol, uint8 srow, uint8 ecol, uint8 erow, uint8 ncol, uint8 nrow);
void writeNum (uint8 n, uint8 col, uint8 row);


//-----------------------------------------------------------------------------

volatile unsigned int ticks;
volatile unsigned int tmsec = 0;
volatile unsigned int tsec = 0;


int main(void)
{
// Ports für SPI usw. setzen
  DDRB  = 0b00101111;
  PORTB = 0b00000000;

  DDRC  = 0b00000110;
  PORTC = 0b00110000;

  DDRD =  0b00110010;
  PORTD = 0b00001000;

  // SPI als master, CLK = F_OSC/2
  SPCR = (1<<SPE)|(1<<MSTR);
  SPSR = (1<<SPI2X);
  // löscht SPIF flag
  SPDR = 0;


  // timer 0 setup
  TCCR0A = (1<<WGM01);				// CTC mode
  TCCR0B |= (1<<CS00)|(1<<CS02); 	// timer0 clock = F_CPU/1024
  OCR0A = 78;						// 10ms
  TIMSK0|=(1<<OCIE0A);				// enable compare_a interrupt

  OledInit();

  OLED_12V_HI;

  delay_ms(100);

  CLS;

  // Interrupts freigeben
  sei();
  srand(1234) ;				// TODO: hier könnte was deutlich Zufälligeres stehen

  //(Taste links)   	(Taste mitte)  		(Taste rechts);
  //(PINC & (1<<PINC4) (PIND & (1<<PIND3) (PINC & (1<<PINC5)

  while(1)
  {
	  Demo();
  }
  return 0;
}


//----------------------------------------------------------------------------------------
//  Demo
//----------------------------------------------------------------------------------------
void Demo(void)
{
	uint8 *r, n, k, l, z, f, p;
	char dk, dl;
	unsigned int a, bumps=0;
	Sprite s[MAXSPRITES];

	write_cmd(0x23);		// enable wrap around for scrolling and copying
	write_cmd(2);

	// lösche Display RAM
	write_cmd(0x15);
	write_cmd(0x00);
	write_cmd(0x3f);
	write_cmd(0x75);
	write_cmd(0x00);
	write_cmd(0x4F);
	for (a = 0; a < (5120); a++)
		write_dat(0x00);

	// Spritepattern in das Display schreiben
	write_cmd(0x15);
	write_cmd(0x07);
	write_cmd(0x13);
	write_cmd(0x75);
	write_cmd(0x00);
	write_cmd(0x4f);
	delay_ms(5);
	r = (uint8*) balls;
	for (a = 0; a < 546; a++) {
		write_dat(pgm_read_byte(r));
		r++;
	}
	// hier die Zahlen
	r = (uint8*) font5x7;
	for (a = 0; a < 273; a++) {
		write_dat(pgm_read_byte(r));
		r++;
	}

	// zum Anschauen
	delay_ms(3000);
	// Spritepattern in den unsichtbaren Bereich kopieren
	doCopy(0x07,0,0x13,0x29,0x3a,0);
	// aufs Copy-Kommando warten
	delay_ms(1);
	// hier die Zahlen
	doCopy(0x07,0x2a,0x13,0x3e,0x3a,0x3b);
	//
	delay_ms(1);
	// clear screen
	CLS;

	// Sprite Kontrollstrukturen intitialisieren
	for (n = 0; n < MAXSPRITES; n++) {
		s[n].dx = 1;
		s[n].dy = 1;
		s[n].op = 1;
		s[n].bcol = pt[n].bcol;
		s[n].brow = pt[n].brow;
	}
	// Startkoordinaten für Sprites und Bewegungsrichtung setzen
	s[0].x = 14; s[0].y = 0;
	s[1].x = 19; s[1].y = 55;
	s[2].x = 95; s[2].y = 29; s[2].dx = -1; s[2].dy = -1;
	s[3].x = 45; s[3].y = 70; s[3].dy = -1;
	s[4].x = 67; s[4].y = 12; s[4].dx = -1;	s[4].dy = -1;
	s[5].x = 51; s[5].y = 21; s[5].dy = -1;
	s[6].x = 61; s[6].y = 48;
	s[7].x = 28; s[7].y = 40;
	k = 14;
	l = 0;
	dk = 1;
	dl = 1;
	z = 0;
	p = 0;
	cli();
	EICRA=0b0100;
	EIMSK=0b10;
	sei();
	ticks=0;
	for (;;) {
		z++;
		// Spritepattern durchrotieren und Sprites darstellen
		for (n = 0; n < MAXSPRITES; n++) {
			if (p == MAXPATTERN - 1) p = 0;
			s[n].scol = pt[p].scol;
			s[n].srow = pt[p].srow;
			s[n].ecol = pt[p].ecol;
			s[n].erow = pt[p].erow;
			// bei ungeraden X-Koordinaten wird ein um ein Pixel verschobenes Pattern
			// auf der selben Position angezeigt
			s[n].po = 0;
			if (s[n].x & 1)
				s[n].po = 7;
			s[n].srow += s[n].po;
			s[n].erow += s[n].po;
			s[n].op = 1;
			doSprite(&s[n]);
			p++;
		}
		delay_ms(15);
		n = MAXSPRITES;
		do {
			n--;
			// Sprites in umgekehrter Reihenfolge löschen
			s[n].op = 0;
			doSprite(&s[n]);
			s[n].srow -= s[n].po;
			s[n].erow -= s[n].po;
			// Update Position
			s[n].x += s[n].dx;
			s[n].y += s[n].dy;
			// am Rand undrehen
			if (s[n].x > 106) { s[n].dx = -1; bumps++; }
			if (s[n].x < 16) { s[n].dx = 1; bumps++; }
			if (s[n].y > 70) { s[n].dy = -1; bumps++; }
			if (s[n].y < 2)  { s[n].dy = 1; bumps++; }
		} while (n > 0);
		// bei jedem z.Frame, Linien/Rechtecke im Hintergrund zeichen
		if (z == 3) {
			// f = Spalte, k = x-Koordinate, l= Reihe
			f = (k >> 1);
			drawRect(f, 0, f, 79, 0, 0);
			drawRect(7, l, 57, l, 0, 0);
			if (k > 112) dk = -1;
			if (k < 16)  dk = 1;
			if (l > 78)  dl = -1;
			if (l < 1)   dl = 1;
			k += dk;
			l += dl;
			f = (k >> 1);
			if (k & 1) drawRect(f, 0, f, 79, 0x04, 0);
			else drawRect(f, 0, f, 79, 0x40, 0);
			drawRect(7, l, 57, l, 0x44, 0);
			z = 0;
		}

		// Bumps einblenden
		if (bumps>999) bumps=0;
		writeNum ((bumps - (bumps % 100))/100, 7, 0);
		writeNum (((bumps%100) - (bumps % 10))/10, 10, 0);
		writeNum (bumps%10, 13, 0);

		// nach TSLEEP Sekunden schlafen gehen
		if ((EICRA & 0b0100) && (ticks>(TSLEEP*100))){
		OLED_12V_LO;
		write_cmd(0xAE);
		set_sleep_mode(SLEEP_MODE_PWR_DOWN);
		PRR=255;
		sleep_mode();
		PRR=0;
		write_cmd(0xAF);
		OLED_12V_HI;
		srand(ticks);
		ticks=0;
		}
	}

}


// Tasteninterrupt
ISR(INT1_vect) {
	srand(ticks);
	ticks=0;
}

// timer0 ISR
ISR (TIMER0_COMPA_vect)
{
	// hier könnte noch mehr passieren
	ticks++;
	tmsec += 10;
	if (tmsec == 1000) {
		tsec++;
		tmsec = 0;
	}
}

// copy digits 0-9 into screen
void writeNum (uint8 n, uint8 col, uint8 row) {
	// digits are 3 colums, 7 rows
	doCopy(digit[n].scol,digit[n].srow,digit[n].scol+2,digit[n].srow+6,col,row);
}


void doSprite(Sprite *sprite) {
	// x = horizontal pixel coordinate
	// y = screen row
	// col = (x/2) = screen column
	// scol = sprite pattern start column, srow = sprite pattern start row
	// ecol = sprite pattern end column, erow = sprite pattern end row
	// bcol = buffer start column, brow = buffer start row
	// op = Operation: 1=set sprite, 0=restore backgroud

	uint8 dcol, drow;

	dcol = sprite->ecol - sprite->scol;	// horizontal size
	drow = sprite->erow - sprite->srow;	// vertical size
	sprite->col = sprite->x >> 1;
	if (sprite->op) {
		// save background to buffer
		write_cmd(0x25);		// enter copy mode
		write_cmd(sprite->col);			// start column
		write_cmd(sprite->y);			// start row
		write_cmd(sprite->col + dcol);	// end column
		write_cmd(sprite->y + drow);	// end row
		write_cmd(sprite->bcol);		// new column
		write_cmd(sprite->brow);		// new row
		// copy sp pattern to screen
		write_cmd(0x25);		// enter copy mode
		write_cmd(sprite->scol);		// start column
		write_cmd(sprite->srow);		// start row
		write_cmd(sprite->scol + dcol);	// end column
		write_cmd(sprite->srow + drow);	// end row
		write_cmd(sprite->col);			// new column
		write_cmd(sprite->y);			// new row
	} else {
		// restore background from buffer
		write_cmd(0x25);		// enter copy mode
		write_cmd(sprite->bcol);		// start column
		write_cmd(sprite->brow);		// start row
		write_cmd(sprite->bcol + dcol);	// end column
		write_cmd(sprite->brow + drow);	// end row
		write_cmd(sprite->col);			// new column
		write_cmd(sprite->y);			// new row
	}
}

void doCopy(uint8 scol, uint8 srow, uint8 ecol, uint8 erow, uint8 ncol, uint8 nrow) {
	write_cmd(0x23); // enable x-wraparound
	write_cmd(2);
	write_cmd(0x25); // enter copy mode
	write_cmd(scol); // start column
	write_cmd(srow); // start row
	write_cmd(ecol); // end column
	write_cmd(erow); // end row
	write_cmd(ncol); // new column
	write_cmd(nrow); // new row
}

void drawRect(uint8 scol, uint8 srow, uint8 ecol, uint8 erow, uint8 pattern, uint8 fill) {
	write_cmd(0x23);		// fill rectangle
	write_cmd(fill & 1);	// or not
	write_cmd(0x24);		// draw rectangle
	write_cmd(scol);		// start column
	write_cmd(srow);		// start row
	write_cmd(ecol);		// end column
	write_cmd(erow);		// end row
	write_cmd(pattern);	// drawing gray scale pattern, left and right pixel
}

//----------------------------------------------------------------------------------------
// OLED-Display initialisieren SSD1325
//----------------------------------------------------------------------------------------
void OledInit(void)
{
  OLED_RES_LO;
  delay_ms(10);
  OLED_RES_HI;
  delay_ms(200);
  //OLED_12V_HI;

  write_cmd(0x15);           // Set Column Address
  write_cmd(0x00);           // Begin  7 Offset! 0x07
  write_cmd(0x3f);           // End   57 	0x39
  write_cmd(0x75);           // Set Row Address
  write_cmd(0x00);           // Begin  0
  write_cmd(0x4F);           // End   79

  write_cmd(0x86);           // Set Current Range 84h:Quarter, 85h:Half, 86h:Full
  write_cmd(0x81);           // Set Contrast Control
  write_cmd(0x2D);
  write_cmd(0xBE);           // Set VCOMH Voltage
  write_cmd(0x11);           // write_cmd(0x11);  default --> bei Streifenfehlern ?

  write_cmd(0xBC);           // Set Precharge Voltage
  write_cmd(0x0F);

  write_cmd(0xA0);           // Set Re-map
  write_cmd(0x41);           //od. 0x43

  write_cmd(0xA6);           // Entire Display OFF, all pixels turns OFF
  write_cmd(0xA8);           // Set Multiplex Ratio
  write_cmd(0x4F);           // multiplex ratio N from 16MUX-80MUX // 80!
  write_cmd(0xB1);           // Set Phase Length
  write_cmd(0x22);
  write_cmd(0xB2);           // Set Row Period
  write_cmd(0x46);
  write_cmd(0xB0);           // Set Pre-charge Compensation Enable
  write_cmd(0x08);
  write_cmd(0xB4);           // Set Pre-charge Compensation Level
  write_cmd(0x00);
  write_cmd(0xB3);           // Set Display Clock
  write_cmd(0xA0);
  write_cmd(0xBF);           // Set Segment Low Voltage (VSL)
  write_cmd(0x0D);

  write_cmd(0xB8);           // Set Gray Scale Table
  write_cmd(0x01);
  write_cmd(0x11);
  write_cmd(0x22);
  write_cmd(0x32);
  write_cmd(0x43);
  write_cmd(0x54);
  write_cmd(0x65);
  write_cmd(0x76);

  write_cmd(0xAE);           // Set Display OFF
  write_cmd(0xAD);           // Set Master Configuration
  write_cmd(0x02);
  delay_ms(75);              // dannach 75ms Pause!? 7365us
  write_cmd(0xA4);           // Normal Display
  write_cmd(0xAF);           // Set Display ON
  write_cmd(0xE3);           //NOP
}


//-----------------------------------------------------------------------------
// send a data byte to oled controller
//-----------------------------------------------------------------------------
void write_dat(uint8 data)
{
  OLED_CS_LO;
  OLED_DC_HI;
  // the display won't work without the 2us delay
  write_byte(data); delay_us(2);
  OLED_CS_HI;
}


//-----------------------------------------------------------------------------
// send a command byte to oled controller
//-----------------------------------------------------------------------------
void write_cmd(uint8 command)
{
  OLED_CS_LO;
  OLED_DC_LO;
  write_byte(command); delay_us(2);
  OLED_DC_HI;

}


//-----------------------------------------------------------------------------
// send a data or command byte
//-----------------------------------------------------------------------------
void write_byte(uint8 dat_or_cmd)
{

#if HW_SPI

  while (!(SPSR & (1<<SPIF)));
  SPDR = dat_or_cmd;

#elif !HW_SPI

  int i;

  for(i=0; i<8; i++)
  {
    if (dat_or_cmd & 0x80) { OLED_SDIN_HI; }
    else                   { OLED_SDIN_LO; }
    dat_or_cmd = dat_or_cmd << 1;
    OLED_SCLK_LO;
    OLED_SCLK_HI;
  }
  OLED_SDIN_LO;

#endif
}

